import re


from svst.config import read_configuration


config = read_configuration()


def is_ignored_file(root: str, file: str) -> bool:
    """Return if the file is ignored by the configuration.

    Args:
        root: Directory path string.
        file: Name of the file.

    Returns:
        Everything that is not a `.py` extension file is ignored.
        Checks if a path prefix, a directory or a file need to be ignored based on the configurations.
    """

    if not file.endswith(".py"):
        return True

    if not config:
        return False

    for prefix_path in (
        [item for item in config["ignore_path_prefix"].split("\n") if item]
        if "ignore_path_prefix" in config
        else []
    ):
        if root.replace("./", "", 1).startswith(prefix_path):
            return True

    for directory in (
        [item for item in config["ignore_directory"].split("\n") if item]
        if "ignore_directory" in config
        else []
    ):
        if directory in root:
            return True

    if (
        file in [item for item in config["ignored_file_names"].split("\n") if item]
        if "ignored_file_names" in config
        else []
    ):
        return True

    return False



import ast

from svst import output


class VariableAnnotationChecker(ast.NodeVisitor):
    def __init__(self, filename):
        self.filename = filename
        self.annotated_vars = set()
        self.scope_stack = []
        self.error_messages = []

    def enter_scope(self, scope_name):
        self.scope_stack.append(scope_name)

    def exit_scope(self):
        self.scope_stack.pop()

    def current_scope(self):
        return ".".join(self.scope_stack)

    def generic_visit(self, node):
        for field, value in ast.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        item.parent = node
                        self.visit(item)
            elif isinstance(value, ast.AST):
                value.parent = node
                self.visit(value)

    def report(self, node, message):
        current_scope = self.current_scope()
        full_scope = ".".join(self.scope_stack)
        self.error_messages.append(
            output.output_structure_constructor(
                self.filename,
                node.lineno,
                message,
                full_scope,
            )
        )

    def check_target(self, node):
        if isinstance(node, (ast.Tuple, ast.List)):
            for el in node.elts:
                self.check_target(el)
        elif isinstance(node, ast.Name):
            if node.id not in self.annotated_vars:
                self.report(node, f'Variable "{node.id}"')
        elif isinstance(node, ast.Assign):
            for target in node.targets:
                self.check_target(target)

    def visit_AnnAssign(self, node):
        if node.simple and node.target and isinstance(node.target, ast.Name):
            self.annotated_vars.add(node.target.id)
        self.generic_visit(node)

    def visit_Assign(self, node):
        for target in node.targets:
            self.check_target(target)
        self.generic_visit(node)

    def visit_AugAssign(self, node):
        self.check_target(node.target)
        self.generic_visit(node)

    def visit_For(self, node):
        self.check_target(node.target)
        self.generic_visit(node)

    def visit_With(self, node):
        for item in node.items:
            if item.optional_vars:
                self.check_target(item.optional_vars)
        self.generic_visit(node)

    def visit_Lambda(self, node):
        # No annotation check for Lambda, as they cannot be type annotated.
        pass

    def visit_arg(self, node):
        parent = getattr(node, "parent", None)
        if isinstance(parent, ast.Lambda) or node.arg == "self":
            return
        if not node.annotation:
            self.report(node, f'Function argument "{node.arg}"')
        else:
            self.annotated_vars.add(node.arg)
        self.generic_visit(node)

    def visit_Comprehension(self, node):
        self.check_target(node.target)
        self.generic_visit(node)

    def visit_ExceptHandler(self, node):
        if node.name and not isinstance(node.name, ast.Name):
            self.report(node, f'Exception handler variable "{node.name}"')
        self.generic_visit(node)

    def visit_Attribute(self, node):
        # Skip checking annotations for attributes
        pass

    def visit_If(self, node):
        self.generic_visit(node)

    def visit_FunctionDef(self, node):
        # Check default values in function definitions
        self.enter_scope(node.name)
        for default in node.args.defaults:
            if isinstance(default, ast.Name):
                self.check_target(default)
        self.generic_visit(node)
        self.exit_scope()

    def visit_Try(self, node):
        # Check variables in the 'except' part
        self.enter_scope("try")
        for handler in node.handlers:
            if handler.name and not isinstance(handler.name, ast.Name):
                self.report(
                    handler,
                    f'Exception handler variable "{handler.name}"',
                )
        self.generic_visit(node)
        self.exit_scope()

    def visit_ClassDef(self, node):
        self.enter_scope(node.name)
        self.generic_visit(node)
        self.exit_scope()

    # Checking List, Set, Dict comprehensions
    def visit_ListComp(self, node):
        for comprehension in node.generators:
            self.visit_Comprehension(comprehension)
        self.generic_visit(node)

    def visit_SetComp(self, node):
        for comprehension in node.generators:
            self.visit_Comprehension(comprehension)
        self.generic_visit(node)

    def visit_DictComp(self, node):
        for comprehension in node.generators:
            self.visit_Comprehension(comprehension)
        self.generic_visit(node)

    # Check variables in while loop
    def visit_While(self, node):
        self.enter_scope("while")
        self.check_target(node.test)
        self.generic_visit(node)
        self.exit_scope()



from typing import Optional
from typing_extensions import TypedDict

import re


class OutputTypedDict(TypedDict):
    file_name: Optional[str]
    line_number: int
    variable_name: str
    variable_scope: str


def output_structure_constructor(
    file_name: Optional[str],
    line_number: int,
    variable_name: str,
    variable_scope: str,
) -> OutputTypedDict:
    """Constructor for dictionary output.

    Args:
        file_name: Name of the file being analysed. None when parse_code is used on bare str.
        line_number: int number of the line where the error occurred.
        variable_name: Name of the variable where the error occurred.
        variable_scope: Scope of the variable where the error occurred.

    Returns:
        OutputTypedDict(TypedDict):
            file_name: Optional[str],
            line_number: int,
            variable_name: str,
            variable_scope: str
    """

    return {
        "file_name": file_name,
        "line_number": line_number,
        "variable_name": variable_name,
        "variable_scope": variable_scope,
    }


def output_string_constructor(
    output_structure: OutputTypedDict,
) -> str:
    """Constructor for string output.

    Args:
        output_structure:
            file_name: Optional[str],
            line_number: int,
            variable_name: str,
            variable_scope: str

    Returns:
        str: {file_name}:{line_number}: error: Variable "{variable_name}" is missing a
        standalone variable type annotation in the scope "{variable_scope}"  [no-untyped-var]
    """

    return (
        f"{output_structure['file_name']}:{output_structure['line_number']}: "
        f"error: {output_structure['variable_name']} is missing a type annotation in the "
        f"scope \"{output_structure['variable_scope']}\"  [no-untyped-var]"
    )


def output_string_converter_terminal(output_string: str):
    """Converts a plain error string into a colored printable one.

    Args:
        output_string: Plain `svst` error output string.

    Returns:
        Colored printable string.
    """

    match = re.search(r"^(\./)?(.+:)([0-9]+:)(.+:)(.+)(\[.+)?$", output_string)
    if not match:
        return output_string

    message: str = f"{match.group(2)}{match.group(3)}\033[31m{match.group(4)}\033[0m{match.group(5)}\033[33m{match.group(6)}\033[0m"

    return message



import ast
import os
import io

from svst import utils, output, parsing

from typing import List, Tuple, Optional

from mypy import api as mypy_api


def parse_code(
    code: str,
    file_name: Optional[str] = None,
) -> List[output.OutputTypedDict]:
    """Parse code directly and return a list of output dictionaries.

    Args:
        code: String of python code.
        file_name: Path of the file being analysed.

    Returns:
        A list of OutputTypedDict that contain the info of the svst error.
    """

    tree = ast.parse(code, file_name)

    checker = parsing.VariableAnnotationChecker(file_name)
    checker.visit(tree)

    return checker.error_messages


def mypy_run(
    path: str,
) -> List[str]:
    """Use mypy api to analyse a path.

    Args:
        path: File or directory path string.

    Yields:
        `mypy` error string.
    """
    results: Tuple[str, str, int] = mypy_api.run([path])

    for error_message in results[0].split("\n")[
        :-2
    ]:  # 2 lines removed at the end to clean the output
        yield error_message


def run(
    path_list: List[str],
    mypy: bool = False,
) -> List[str]:
    """Run path list.

    Args:
        path_list: List of path strings.
        mypy: Run mypy for the same `path_list`.

    Yields:
        str: {file_name}:{line_number}: error: Variable "{variable_name}" is missing a
        standalone variable type annotation in the scope "{variable_scope}"  [no-untyped-var]
    """

    path: str
    for path in path_list:
        root: str
        dirs: List[str]
        files: List[str]
        for root, dirs, files in os.walk(path):
            file: str
            for file in files:
                if utils.is_ignored_file(root, file):
                    continue

                file_path: str = os.path.join(root, file)

                file_buffer: io.TextIOWrapper
                with open(file_path, "r") as file_buffer:
                    svst_errors = parse_code(file_buffer.read(), file_path)

                    for svst_error in svst_errors:
                        yield output.output_string_constructor(svst_error)

        if mypy:
            for line in mypy_run(path):
                yield line
